import sys
import os
import re
import contextlib
import tempfile
import importlib.resources
import json

import pysam
import pandas as pd

import handygenome
import handygenome.tools as tools
import handygenome.network as network
import handygenome.logutils as logutils
#import handygenome.publicdb.ncbi_postprocess as ncbi_postprocess
#import handygenome.fastahandler as fastahandler


URL_AUTHORITY = 'ftp.ncbi.nlm.nih.gov'
URL_PATH_REFSEQ_GENOME = '/genomes/refseq'
URL_PATH_DBSNP = "/snp/latest_release/VCF"
REFSEQ_ACC_PATSTR = r'(?P<refseq_acc>(?P<refseq_acc_prefix>[^_]+)_(?P<refseq_acc_main>[^_]+)\.(?P<refseq_acc_version>[^_]+))'
REFVER_PATSTR = r'(?P<refver>(?P<refver_main>[^.]+)(\.(?P<refver_sub>[^.]+))?)'
GENOME_DIR_PAT = re.compile(REFSEQ_ACC_PATSTR + '_' + REFVER_PATSTR)


#SPECIES_PARDIR_MAP = {
#    'Homo_sapiens': 'vertebrate_mammalian',
#    'Mus_musculus': 'vertebrate_mammalian',
#    'Musa_acuminata': 'plant',
#}


class UnavailableSpeciesError(Exception):
    pass


class UnavailableRefverError(Exception):
    pass


#################################
# RefSeq ftp genome directories #
#################################

class RefseqGenomePaths:
    """Data structure:
        :species_toppaths: dict(species superset(e.g. vertebrate_mammalian) -> dict(species -> path))
        :refver_toppaths: dict(species -> dict(refver -> dict(accesion -> path)))
        :latest_refver_toppaths: dict(species -> dict(refver -> path))
        :refver_details: dict(refver -> dict(file type(e.g. assembly_report, fasta) -> file path))

    Data update methods
    * get_species_toppath (main updater)
    """

    # cache managers
    cache_path = os.path.join(
        handygenome.DATA_DIR, 
        'refseq_genome_toppaths.json',
    )

    def remove_cache(self):
        os.remove(self.__class__.cache_path)

    def load_cache(self):
        #with open(self.__class__.cache_path, 'rt') as f:
        #    self.species_toppaths = json.load(f)
        with open(self.__class__.cache_path, 'rt') as f:
            cacheobj = json.load(f)

        self.species_toppaths = cacheobj['species_toppaths']
        self.refver_toppaths = cacheobj['refver_toppaths']
        self.latest_refver_toppaths = cacheobj['latest_refver_toppaths']
        self.refver_details = cacheobj['refver_details']

    def save_cache(self):
        cacheobj = dict()
        cacheobj['species_toppaths'] = self.species_toppaths
        cacheobj['refver_toppaths'] = self.refver_toppaths
        cacheobj['latest_refver_toppaths'] = self.latest_refver_toppaths
        cacheobj['refver_details'] = self.refver_details

        with open(self.__class__.cache_path, 'wt') as f:
            json.dump(cacheobj, f)

    def init_path_dicts(self):
        if os.path.exists(self.__class__.cache_path):
            self.load_cache()
        else:
            self.species_toppaths = dict()
            self.refver_toppaths = dict()
            self.latest_refver_toppaths = dict()
            self.refver_details = dict()

    ####

    all_genome_supersets = [
        #'assembly_summary_refseq.txt',
        #'assembly_summary_refseq_historical.txt',
        #'README.txt',

        'vertebrate_mammalian',
        'vertebrate_other',
        'plant',
        'bacteria',
        'viral',
        'fungi',
        'invertebrate',
        'protozoa',
        'archaea',

        #'unknown',
        #'metagenomes',
    ]

    def __init__(self, force_update=False):
        self.initialize(force_update=force_update)

    def __del__(self):
        self.save_cache()

    def initialize(self, force_update=False):
        if force_update:
            self.remove_cache()
        self.init_path_dicts()

    #####

    def search_for_species_toppath(self, query_species):
        for superset, subdic in self.species_toppaths.items():
            for species in subdic.keys():
                if species == query_species:
                    return subdic[species]
        return None

    def get_species_toppath(self, species):
        toppath = self.search_for_species_toppath(species)
        if toppath is None:  # now query species is not in the data
            for species_superset in self.__class__.all_genome_supersets:
                success = self.set_species_toppaths(
                    species_superset=species_superset,
                    retry_count=None, 
                    retry_interval=5, 
                    timeout=180,  # set a large value for "bacteria"
                )
                toppath = self.search_for_species_toppath(species)
                if toppath is None:
                    continue
                else:
                    break

        if toppath is None:
            raise UnavailableSpeciesError(f'Species "{species}" could not be found from the list of RefSeq genomes')

        return toppath

    def set_species_toppaths(
        self, species_superset,
        retry_count=None, retry_interval=5, timeout=60,
    ):
        if species_superset in self.species_toppaths.keys():
            success = True
        else:
            superset_path = network.join_ftp_paths(URL_PATH_REFSEQ_GENOME, species_superset)
            logutils.print_timestamp(f'Fetching from "{superset_path}"')
            try:
                nlst_result = network.ftp_nlst(
                    URL_AUTHORITY, superset_path,
                    retry_count=retry_count, 
                    retry_interval=retry_interval, 
                    timeout=timeout,
                    verbose=True,
                )
            except network.MaxRetryError as exc:
                logutils.print_timestamp(
                    f'Skipping "{superset_path}" because maximum retry count was exceeded'
                )
                success = False
            else:
                # do update
                self.species_toppaths[species_superset] = dict()
                for species_path in nlst_result:
                    species = species_path.split('/')[-1]
                    self.species_toppaths[species_superset][species] = species_path
                success = True
                self.save_cache()

        return success

    #####

    def set_refver_toppaths(
        self, species,
        retry_count=None, retry_interval=5, timeout=60,
    ):
        species_toppath = self.get_species_toppath(species)
        nlst_path = network.join_ftp_paths(species_toppath, 'all_assembly_versions')

        fname_list = network.ftp_nlst(
            URL_AUTHORITY, nlst_path,
            retry_count=retry_count, 
            retry_interval=retry_interval, 
            timeout=timeout,
            verbose=True,
        )

        self.refver_toppaths[species] = dict()
        for path in fname_list:
            mat = GENOME_DIR_PAT.fullmatch(os.path.basename(path))
            if mat is None:
                continue
            refver = mat.group('refver_main')
            acc_ver = int(mat.group('refseq_acc_version'))
            self.refver_toppaths[species].setdefault(refver, dict())
            self.refver_toppaths[species][refver][acc_ver] = path

        self.save_cache()

    def set_latest_refver_toppaths(
        self, species,
        retry_count=None, retry_interval=5, timeout=60,
    ):
        self.set_refver_toppaths(
            species,
            retry_count=retry_count, 
            retry_interval=retry_interval, 
            timeout=timeout,
        )

        self.latest_refver_toppaths[species] = dict() 
        for refver, subdic in self.refver_toppaths[species].items():
            self.latest_refver_toppaths[species][refver] = max(subdic.items(), key=lambda x: x[0])[1]

        self.save_cache()

    def get_refver_toppath(
        self, refver, species,
        retry_count=None, retry_interval=5, timeout=60,
    ):
        """Returns: "/~/~/all_assembly_versions"
        """
        if (species not in self.latest_refver_toppaths):
            self.set_latest_refver_toppaths(
                species, 
                retry_count=retry_count, 
                retry_interval=retry_interval, 
                timeout=timeout,
            )
        if refver not in self.latest_refver_toppaths[species]:
            raise Exception(f'There is no refver named "{refver}" in the species "{species}"')
        return self.latest_refver_toppaths[species][refver]

    def get_latest_refver_toppaths(
        self, species,
        retry_count=None, retry_interval=5, timeout=60,
    ):
        if species not in self.latest_refver_toppaths:
            self.set_latest_refver_toppaths(
                species, 
                retry_count=retry_count, 
                retry_interval=retry_interval, 
                timeout=timeout,
            )
        return self.latest_refver_toppaths[species]

    #####

    def set_refver_details(
        self, refver, species,
        retry_count=None, retry_interval=5, timeout=60,
    ):
        refver_toppath = self.get_refver_toppath(
            refver, species,
            retry_count=retry_count, 
            retry_interval=retry_interval, 
            timeout=timeout,
        )
        data = dict()
        for path in network.ftp_nlst(
            URL_AUTHORITY, refver_toppath,
            retry_count=retry_count, 
            retry_interval=retry_interval, 
            timeout=timeout,
            verbose=True,
        ):
            if path.endswith('assembly_report.txt'):
                data['assembly_report'] = path
            elif path.endswith('assembly_regions.txt'):
                data['assembly_regions'] = path
            elif (
                path.endswith('genomic.fna.gz')
                and (not path.endswith('rna_from_genomic.fna.gz'))
            ):
                data['fasta'] = path
            elif path.endswith('genomic.gff.gz'):
                data['gff'] = path
            elif path.endswith('protein.faa.gz'):
                data['protein_fasta'] = path
            elif path.endswith('rm.out.gz'):
                data['repeatmasker'] = path

        self.refver_details[refver] = data

        self.save_cache()

    def get_refver_details(
        self, refver, species,
        retry_count=None, retry_interval=5, timeout=60,
        force_update=False,
    ):
        if (refver not in self.refver_details):
            self.set_refver_details(
                refver, species,
                retry_count=retry_count, 
                retry_interval=retry_interval, 
                timeout=timeout,
            )
        return self.refver_details[refver]

    ###############
    # url getters #
    ###############

    def get_url(self, refver, species, key, force_update):
        if force_update:
            self.initialize(force_update=True)
        refver_details = self.get_refver_details(refver, species, force_update=force_update)
        path = refver_details[key]
        return f'https://{URL_AUTHORITY}{path}'

    def get_fasta_url(self, refver, species, force_update=False):
        return self.get_url(refver, species, 'fasta', force_update)

    def get_assembly_report_url(self, refver, species, force_update=False):
        return self.get_url(refver, species, 'assembly_report', force_update)

    def get_assembly_regions_url(self, refver, species, force_update=False):
        return self.get_url(refver, species, 'assembly_regions', force_update)

    def get_gff_url(self, refver, species, force_update=False):
        return self.get_url(refver, species, 'gff', force_update)

    def get_repeatmasker_url(self, refver, species, force_update=False):
        return self.get_url(refver, species, 'repeatmasker', force_update)

    #####################################
    # species/refver availability check #
    #####################################

    def check_species_availability(self, species):
        try:
            toppath = self.get_species_toppath(species)
        except UnavailableSpeciesError:
            return False
        else:
            return True

    def get_available_refvers(
        self, species,
        retry_count=None, retry_interval=5, timeout=60,
    ):
        refver_toppaths_for_species = self.get_latest_refver_toppaths(
            species,
            retry_count=retry_count, 
            retry_interval=retry_interval, 
            timeout=timeout,
        )
        return tuple(refver_toppaths_for_species.keys())

    # find species from a RefSeq refver
    def find_species_without_update(self, RefSeq_refver):
        for species, subdic in self.latest_refver_toppaths.items():
            if RefSeq_refver in subdic.keys():
                return species
        return None

    def find_species(self, RefSeq_refver):
        species = self.find_species_without_update(RefSeq_refver)
        if species is None:
            for species_superset in self.__class__.all_genome_supersets:
                success = self.set_species_toppaths(
                    species_superset=species_superset,
                    retry_count=None, 
                    retry_interval=5, 
                    timeout=180,  # set a large value for "bacteria"
                )
                species = self.find_species_without_update(RefSeq_refver)
                if species is None:
                    continue
                else:
                    break

        if species is None:
            raise UnavailableRefverError(f'Reference version "{RefSeq_refver}" could not be found from the list of RefSeq genomes')

        return species


def reset_genome_paths(force_update=False):
    setattr(sys.modules[__name__], 'GENOME_PATHS', RefseqGenomePaths(force_update=force_update))

reset_genome_paths()
#GENOME_PATHS = RefseqGenomePaths()



################################
# assembly_regions file parser #
################################

def parse_assembly_regions(filepath):
    # get column names and the number of rows to skip
    skiprows = 0
    with open(filepath) as infile:
        for line in infile:
            skiprows += 1
            if line.startswith('# Region-Name'):
                columns = re.sub('^#\s*', '', line).strip().split() 
                break

    # parse the rest with pandas
    result = pd.read_csv(
        filepath, sep='\s+', names=columns, header=None, skiprows=skiprows,
        dtype={'Chromosome-Start': int, 'Chromosome-Stop': int},
    )
    return result


# convenience functions

#def get_fasta_url(refver, species, force_update=False):
#    return GENOME_PATHS.get_fasta_url(refver, species, force_update)
#
#
#def get_assembly_report_url(refver, species, force_update=False):
#    return GENOME_PATHS.get_assembly_report_url(refver, species, force_update)
#
#
#def get_gff_url(refver, species, force_update=False):
#    return GENOME_PATHS.get_gff_url(refver, species, force_update)


########################################
# RefSeq file downloaders and fetchers #
########################################

#def refseqfile_cacher_base(
#    topdir, standard_refver, species, path_basename, force_download, get_url_key,
#    download_msg=None,
#):
#    species_subdir = os.path.join(topdir, species)
#    os.makedirs(species_subdir, exist_ok=True)
#    filepath = os.path.join(species_subdir, path_basename)
#
#    if (not os.path.exists(filepath)) or force_download:
#        did_download = True
#        if download_msg is None:
#            download_msg = (
#                f'Downloading {get_url_key} file for reference version "{standard_refver}"'
#            )
#        logutils.print_timestamp(download_msg)
#
#        url = GENOME_PATHS.get_url(
#            refver=standard_refver, 
#            species=species, 
#            key=get_url_key, 
#            force_update=False,
#        )
#        network.download(url, filepath)
#    else:
#        did_download = False
#
#    return filepath, did_download
#
#
## fasta
#
#FASTAFILE_DIR = os.path.join(handygenome.DATADIR, 'refseq_fasta')
#os.makedirs(FASTAFILE_DIR, exist_ok=True)
#
#def get_unedited_fasta_path(standard_refver, species, force_download=False):
#    filepath, did_download = refseqfile_cacher_base(
#        topdir=FASTAFILE_DIR, 
#        standard_refver=standard_refver, 
#        species=species, 
#        path_basename=f'{standard_refver}.unedited.fna.gz', 
#        force_download=force_download, 
#        get_url_key='fasta',
#    )
#
#    return filepath
#
#
#def get_edited_fasta_path(standard_refver, species, force_download=False):
#    edited_fasta_path = os.path.join(FASTAFILE_DIR, f'{standard_refver}.edited.fasta.gz')
#    if not os.path.exists(edited_fasta_path):
#        unedited_fasta_path = get_unedited_fasta_path(
#            standard_refver, species, force_download=force_download,
#        )
#        from handygenome.publicdb.ncbi_postprocess import postprocess_fasta
#        postprocess_fasta(unedited_fasta_path, edited_fasta_path)
#    return edited_fasta_path
#
#
## assemblyspec
#
#ASSEMBLYFILE_DIR = os.path.join(handygenome.DATADIR, 'assembly_reports')
#os.makedirs(ASSEMBLYFILE_DIR, exist_ok=True)
#
#def get_assemblyfile_path(standard_refver, species, force_download=False):
#    filepath, did_download = refseqfile_cacher_base(
#        topdir=ASSEMBLYFILE_DIR, 
#        standard_refver=standard_refver, 
#        species=species, 
#        path_basename=f'{standard_refver}.txt', 
#        force_download=force_download, 
#        get_url_key='assembly_report',
#    )
#
#    return filepath
#
#
## repeatmasker
#
#REPEATMASKER_DIR = os.path.join(handygenome.DATADIR, 'refseq_repeatmasker')
#os.makedirs(REPEATMASKER_DIR, exist_ok=True)
#
#def get_repeatmasker_path(standard_refver, species, force_download=False):
#    filepath, did_download = refseqfile_cacher_base(
#        topdir=REPEATMASKER_DIR, 
#        standard_refver=standard_refver, 
#        species=species, 
#        path_basename=f'{standard_refver}.rm.out.gz', 
#        force_download=force_download, 
#        get_url_key='repeatmasker',
#    )
#
#    return filepath
#
#
## geneset
#
#GENESET_DIR = os.path.join(handygenome.DATADIR, 'refseq_geneset')
#os.makedirs(GENESET_DIR, exist_ok=True)
#
#def get_geneset_path(standard_refver, species, force_download=False):
#    filepath, did_download = refseqfile_cacher_base(
#        topdir=GENESET_DIR, 
#        standard_refver=standard_refver, 
#        species=species, 
#        path_basename=f'{standard_refver}.unedited.gff.gz', 
#        force_download=force_download, 
#        get_url_key='gff',
#    )
#
#    return filepath





#########################################


#def genome_path_sorting(
#    species_genome_topdir,
#    retry_count=None, retry_interval=5, timeout=60,
#):
#    fname_list = network.ftp_nlst(
#        URL_AUTHORITY, species_genome_topdir,
#        retry_count=retry_count, 
#        retry_interval=retry_interval, 
#        timeout=timeout,
#    )
#
#    genome_path_dict = dict()
#    for path in fname_list:
#        mat = GENOME_DIR_PAT.fullmatch(os.path.basename(path))
#        if mat is None:
#            continue
#
#        refver = mat.group('refver_main')
#        acc_ver = int(mat.group('refseq_acc_version'))
#        genome_path_dict.setdefault(refver, dict())
#        genome_path_dict[refver][acc_ver] = path
#
#    return genome_path_dict
#
#
#def refver_subver_key(refver_sub):
#    if refver_sub is None:
#        refver_sub_key = -1
#    elif refver_sub.startswith('p'):
#        refver_sub_key = int(refver_sub[1:])
#    else:
#        refver_sub_key = int(refver_sub)
#    return refver_sub_key
#
#
#def pick_latest_genome_path(genome_path_dict):
#    result = dict()
#    for refver, subdic in genome_path_dict.items():
#        result[refver] = max(subdic.items(), key=lambda x: x[0])[1]
#    return result
#
#
#def collect_latest_genome_paths(
#    species_list, 
#    ftp=None, 
#    retry_count=10, 
#    retry_interval=1, 
#    timeout=5,
#):
#    def helper(ftp, species_genome_topdir, result):
#        fname_dict = genome_path_sorting(ftp, species_genome_topdir)
#        latest_genome_paths_part = pick_latest_genome_path(fname_dict)
#        result.update(latest_genome_paths_part)
#
#    # access ftp server and collect file paths
#    result = dict()
#    if ftp is None:
#        ftp = network.ftp_login(
#            URL_AUTHORITY, 
#            retry_count=retry_count, 
#            retry_interval=retry_interval, 
#            timeout=timeout,
#        )
#
#    with contextlib.redirect_stdout('/dev/null'):
#        species_pardir_pairs = [
#            (species, SPECIES_PARDIR_MAP[species])
#            for species in species_list
#        ]
#        for species, pardir in species_pardir_pairs:
#            topdir = URL_PATH_REFSEQ_GENOME + f'/{pardir}/{species}/all_assembly_versions'
#            helper(ftp, topdir, result)
#
#    return result
#
#
#def find_assemblyfile_from_genomepath(ftp, genome_path):
#    fname_list = network.ftp_listdir(ftp, genome_path)
#    results = [x for x in fname_list if x.endswith('assembly_report.txt')]
#    if len(results) != 1:
#        raise Exception(f'Cannot find a unique assembly report file')
#    return results[0]
#
#
#def collect_assemblyfile_urls(retry_count=10, retry_interval=1, timeout=5):
#    ftp = network.ftp_login(
#        URL_AUTHORITY, 
#        retry_count=retry_count, 
#        retry_interval=retry_interval, 
#        timeout=timeout,
#    )
#    latest_genome_paths = collect_latest_genome_paths(ftp=ftp)
#    assemblyfile_paths = {
#        refver: find_assemblyfile_from_genomepath(ftp, genome_path)
#        for refver, genome_path in latest_genome_paths.items()
#    }
#
#    # add prefix
#    assemblyfile_urls = {
#        key: f'https://{URL_AUTHORITY}' + val
#        for key, val in assemblyfile_paths.items()
#    }
#
#    # quit ftp
#    ftp.quit()
#
#    return assemblyfile_urls


class RefseqDbsnpPaths:
    pass


def get_dbsnp_urls(retry_count=10, retry_interval=1, timeout=5):
    # access ftp server and collect file paths
    ftp = network.ftp_login(
        URL_AUTHORITY, 
        retry_count=retry_count, 
        retry_interval=retry_interval, 
        timeout=timeout,
    )
    with contextlib.redirect_stdout('/dev/null'):
        fname_list = network.ftp_listdir(ftp, URL_PATH_DBSNP)
        ftp.quit()

    # make result
    pat = re.compile(r'(.+)\.gz(.*)')
    url_dict = dict()
    for fname in fname_list:
        mat = pat.fullmatch(os.path.basename(fname))
        if mat is None:  # CHECKSUM
            continue

        key = mat.group(1)
        url_dict.setdefault(key, dict())

        if mat.group(2) == '':
            subkey = 'vcf'
        elif mat.group(2) == '.md5':
            subkey = 'vcf_md5'
        else:
            continue

        url_dict[key][subkey] = f'https://{URL_AUTHORITY}' + fname

    return url_dict




import re
import collections
import itertools

import Bio.Align
import numpy as np

import handygenome.refgenome.refgenome as refgenome
import handygenome.tools as tools
import handygenome.variant.vcfspec as libvcfspec


##########################
# preset aligner objects #
##########################

ALIGNER_BLASTN = Bio.Align.PairwiseAligner(
    match_score=2,
    mismatch_score=-3,
    query_internal_open_gap_score=-7,
    query_internal_extend_gap_score=-2,
    target_internal_open_gap_score=-7,
    target_internal_extend_gap_score=-2,
)

ALIGNER_EQUAL_MM_GAP = Bio.Align.PairwiseAligner(
    mode='global',
    match_score=3,
    mismatch_score=-3,
    query_internal_open_gap_score=-3,
    query_internal_extend_gap_score=0,
    target_internal_open_gap_score=-3,
    target_internal_extend_gap_score=0,
)


######################
# CIGAR-RELATED ONES #
######################

CIGAROPDICT_ITEMS = [
    ("M", 0),
    ("I", 1),
    ("D", 2),
    ("N", 3),
    ("S", 4),
    ("H", 5),
    ("P", 6),
    ("=", 7),
    ("X", 8),
    ("B", 9),
]
CIGAROPDICT = dict(CIGAROPDICT_ITEMS)
CIGAROPDICT_REV = dict((x[1], x[0]) for x in CIGAROPDICT_ITEMS)
CIGARPAT = re.compile(f'([0-9]+)([{"".join(CIGAROPDICT.keys())}])')
CIGAR_WALK_DICT = { # (target, query)
    0: (True, True),
    1: (False, True),
    2: (True, False),
    3: (True, False),
    4: (False, True),
    5: (False, False),
    6: (False, False),
    7: (True, True),
    8: (True, True),
}
CIGAROPS_BOTH = set(
    key for key, (target, query) in CIGAR_WALK_DICT.items()
    if target and query
)
CIGAROPS_TARGETONLY = set(
    key for key, (target, query) in CIGAR_WALK_DICT.items()
    if target and (not query)
)
CIGAROPS_QUERYONLY = set(
    key for key, (target, query) in CIGAR_WALK_DICT.items()
    if (not target) and query
)
#CIGAROPS_BOTH = {0, 7, 8}
#CIGAROPS_TARGETONLY = {2, 3}
#CIGAROPS_QUERYONLY = {1, 4}


class Cigar:
    # constructors #
    @classmethod
    def from_cigarstring(cls, cigarstring):
        result = cls()
        result.cigartuples = cigarstring_to_cigartuples(cigarstring)
        return result
    
    @classmethod
    def from_cigartuples(cls, cigartuples):
        result = cls()
        result.cigartuples = cigartuples
        return result
    ################

    @property
    def cigarstring(self):
        return cigartuples_to_cigarstring(self.cigartuples)

    def iter_trailing_queryonly(self):
        return iter_trailing_queryonly(self.cigartuples)

    def iter_leading_queryonly(self):
        return iter_leading_queryonly(self.cigartuples)

    def check_DN_outside(self):
        # cigar I may be outside!
        opstrings_rmclip = [x for x in self.opstrings if x not in "SH"]

        return opstrings_rmclip[0] in "DN" or opstrings_rmclip[-1] in "DN"

    def check_clip_inside(self):
        if len(self.opstrings) <= 2:
            return False
        else:
            return not {"S", "H"}.isdisjoint(self.opstrings[1:-1])

    def check_unexpected_pattern(self):
        return (
            self.check_DN_outside()
            or self.check_clip_inside()
            or (not {"B", "P"}.isdisjoint(self.opstrings))
        )


def cigarstring_to_cigartuples(cigarstring):
    result = list()
    for (count, opstring) in CIGARPAT.findall(cigarstring):
        opcode = CIGAROPDICT[opstring]
        count = int(count)
        result.append((opcode, count))
    return result


def cigartuples_to_cigarstring(cigartuples):
    buffer = list()
    for opcode, count in cigartuples:
        opstring = CIGAROPDICT_REV[opcode]
        buffer.append(str(count))
        buffer.append(opstring)
    return "".join(buffer)


def cigartuples_to_cigarstring_spread(cigartuples):
    buffer = list()
    for opcode, count in cigartuples:
        opstring = CIGAROPDICT_REV[opcode]
        buffer.extend(itertools.repeat(opstring, count))
    return "".join(buffer)


def iter_trailing_queryonly(cigartuples):
    for opcode, count in reversed(cigartuples):
        if opcode in CIGAROPS_QUERYONLY:
            yield opcode, count
        else:
            break


def iter_leading_queryonly(cigartuples):
    for opcode, count in cigartuples:
        if opcode in CIGAROPS_QUERYONLY:
            yield opcode, count
        else:
            break


def get_query_length(cigartuples):
    result = 0
    for opcode, count in cigartuples:
        consume_target, consume_query = CIGAR_WALK_DICT[opcode]
        if consume_query:
            result += count
    return result


def get_target_length(cigartuples):
    result = 0
    for opcode, count in cigartuples:
        consume_target, consume_query = CIGAR_WALK_DICT[opcode]
        if consume_target:
            result += count
    return result


# walk cigar #

class WalkBase:
    def check_targetonly(self):
        return (len(self.target) != 0) and (len(self.query) == 0)

    def check_queryonly(self):
        return (len(self.query) != 0) and (len(self.target) == 0)

    def check_both(self):
        return (len(self.query) != 0) and (len(self.target) != 0)

    def get_target_slice(self):
        return slice(self.target.start, self.target.stop)

    def get_query_slice(self):
        return slice(self.query.start, self.query.stop)

    @property
    def range0(self):
        if self.check_targetonly():
            return self.target
        else:
            return self.query

    def __len__(self):
        return len(self.range0)


class Walk(
    WalkBase,
    collections.namedtuple('Walk', ('target', 'query')), 
):
    pass


class CigarWalk(
    WalkBase,
    collections.namedtuple('CigarWalk', ('cigartup', 'target', 'query')),
):
    pass


def walk_cigar(cigartuples, target_start0, with_cigartup=False):
    query_start0 = 0

    target_end0 = target_start0
    query_end0 = query_start0

    for cigartup in cigartuples:
        opcode, count = cigartup
        consume_target, consume_query = CIGAR_WALK_DICT[opcode]
        if consume_target:
            target_end0 += count
        if consume_query:
            query_end0 += count

        target_range0 = range(target_start0, target_end0)
        query_range0 = range(query_start0, query_end0)

        #yield Walk(target_range0, query_range0)
        if with_cigartup:
            yield CigarWalk(cigartup, target_range0, query_range0)
        else:
            yield Walk(target_range0, query_range0)

        target_start0 = target_end0
        query_start0 = query_end0


def walk_cigar_targetonly(cigartuples, target_start0):
    #query_start0 = 0

    target_end0 = target_start0
    #query_end0 = query_start0

    for cigartup in cigartuples:
        opcode, count = cigartup
        consume_target, consume_query = CIGAR_WALK_DICT[opcode]
        if consume_target:
            target_end0 += count
        #if consume_query:
        #    query_end0 += count

        target_range0 = range(target_start0, target_end0)
        #query_range0 = range(query_start0, query_end0)

        yield target_range0

        target_start0 = target_end0
        #query_start0 = query_end0


# split cigar #

def split_cigar(
    cigartuples, reference_start0, split_range0,
    trailing_queryonly_to_right=False,
):
    """Position of queryonly(I, S) cigarop is considered to be on the right,
    except for those on the right end, which is controlled by 
    "trailing_queryonly_to_right" argument.

    Args:
        - trailing_queryonly_to_right:
            If True, softclips or insertions on the right end are regarded as 
            residing in the position on the right. 
    """
    cigarwalks_grouped = itertools.groupby(
        walk_cigar(cigartuples, reference_start0, with_cigartup=True),
        key=(
            lambda cigarwalk: (
                (len(cigarwalk.target) == 0) and 
                (len(cigarwalk.query) != 0)
            )
        )
    )
    cigarwalks_grouped = [(x[0], list(x[1])) for x in cigarwalks_grouped]

    if cigarwalks_grouped[-1][0]:  # there exists trailing queryonly
        cigarwalks_trailing_queryonly = cigarwalks_grouped[-1][1]  # list of CigarWalk
        cigarwalks_others = cigarwalks_grouped[:-1]  # list of (key, [Cigarwalk, ...])
    else:
        cigarwalks_trailing_queryonly = None
        cigarwalks_others = cigarwalks_grouped

    # main
    tuples_before = list()
    tuples_within = list()
    tuples_after = list()
    # handle non-trailing-queryonly
    for is_queryonly, cigarwalk_list in cigarwalks_others:
        if is_queryonly:  # queryonly
            for cigarwalk in cigarwalk_list:
                before_portion, within_portion, after_portion = cigarsplit_helper_queryonly(
                    cigarwalk, split_range0, is_trailing=False,
                )
                cigarsplit_helper_append(
                    before_portion, within_portion, after_portion,
                    tuples_before, tuples_within, tuples_after,
                )
        else:
            for cigarwalk in cigarwalk_list:
                before_portion, within_portion, after_portion = cigarsplit_helper_non_queryonly(
                    cigarwalk, split_range0,
                )
                cigarsplit_helper_append(
                    before_portion, within_portion, after_portion,
                    tuples_before, tuples_within, tuples_after,
                )

    # handle trailing queryonly
    if cigarwalks_trailing_queryonly is not None:
        for cigarwalk in cigarwalks_trailing_queryonly:
            before_portion, within_portion, after_portion = cigarsplit_helper_queryonly(
                cigarwalk, split_range0, is_trailing=(not trailing_queryonly_to_right),
            )
            cigarsplit_helper_append(
                before_portion, within_portion, after_portion,
                tuples_before, tuples_within, tuples_after,
            )

    return tuples_before, tuples_within, tuples_after


def cigarsplit_helper_non_queryonly(cigarwalk, split_range0):
    offset_split_start = split_range0.start - cigarwalk.target.start
    offset_split_end = cigarwalk.target.stop - split_range0.stop
    # before
    if offset_split_start > 0:
        #before_portion_length = min(offset_split_start, len(cigarwalk.target))
        before_portion = (
            cigarwalk.cigartup[0], 
            min(offset_split_start, len(cigarwalk.target)),
        )
    else:
        before_portion = None
    # after
    if offset_split_end > 0:
        after_portion = (
            cigarwalk.cigartup[0], 
            min(offset_split_end, len(cigarwalk.target)),
        )
    else:
        after_portion = None
    # within
    if (
        cigarwalk.target.start < split_range0.stop and
        cigarwalk.target.stop > split_range0.start
    ):
        cigarop_idx_start0 = max(0, offset_split_start)
        cigarop_idx_end0 = len(cigarwalk.target) - max(0, offset_split_end)
        within_portion = (
            cigarwalk.cigartup[0],
            cigarop_idx_end0 - cigarop_idx_start0
        )
    else:
        within_portion = None

    return before_portion, within_portion, after_portion


def cigarsplit_helper_queryonly(cigarwalk, split_range0, is_trailing):
    if is_trailing:  # assign to the position on the left
        pos0 = cigarwalk.target.start - 1
    else:
        pos0 = cigarwalk.target.start

    if pos0 in split_range0:
        before_portion = None
        within_portion = cigarwalk.cigartup
        after_portion = None
    elif pos0 < split_range0.start:
        before_portion = cigarwalk.cigartup
        within_portion = None
        after_portion = None
    elif pos0 >= split_range0.stop:
        before_portion = None
        within_portion = None
        after_portion = cigarwalk.cigartup

    return before_portion, within_portion, after_portion


def cigarsplit_helper_append(
    before_portion, within_portion, after_portion,
    tuples_before, tuples_within, tuples_after,
):
    if before_portion is not None:
        tuples_before.append(before_portion)
    if within_portion is not None:
        tuples_within.append(within_portion)
    if after_portion is not None:
        tuples_after.append(after_portion)


# old versions of cigar split functions #

def split_cigar_old(cigartuples, reference_start0, split_range0):
    """This function also returns cigartuples within split_range0"""
    tuples_before = list()
    tuples_within = list()
    tuples_after = list()

    for cigarunit_tup, target_range0 in zip(
        cigartuples, 
        walk_cigar_targetonly(cigartuples, reference_start0)
    ):
        if len(target_range0) == 0:
            # query-only cigar operation
            # start and stop of target_range0 are the same
            # start of target_range0 indicates the position next to the insertion site
            pos0 = target_range0.start
            if pos0 < split_range0.start:
                tuples_before.append(cigarunit_tup)
            elif pos0 > split_range0.stop:
                tuples_after.append(cigarunit_tup)
            else:
                # query-only cigarops immediately before or after "split_range0" are all included in "tuples_within"
                tuples_within.append(cigarunit_tup)
        else:
            # current cigarwalk entirely before split_range0
            if target_range0.stop <= split_range0.start:
                tuples_before.append(cigarunit_tup)
            # current cigarwalk entirely after split_range0
            elif target_range0.start >= split_range0.stop:
                tuples_after.append(cigarunit_tup)
            # current cigarwalk overlaps split_range0
            else:
                # add to tuples_before
                if target_range0.start < split_range0.start:
                    tuple_to_add = (
                        cigarunit_tup[0], 
                        split_range0.start - target_range0.start,
                    )
                    tuples_before.append(tuple_to_add)
                # add to tuples_after
                if target_range0.stop > split_range0.stop:
                    tuple_to_add = (
                        cigarunit_tup[0], 
                        target_range0.stop - split_range0.stop,
                    )
                    tuples_after.append(tuple_to_add)
                # add to tuples_within
                tuple_to_add = (
                    cigarunit_tup[0], 
                    (
                        min(target_range0.stop, split_range0.stop)
                        - max(target_range0.start, split_range0.start)
                    )
                )
                tuples_within.append(tuple_to_add)

    return tuples_before, tuples_within, tuples_after


def get_cigars_before_region(cigartuples, reference_start0, reference_end0, split_range0):
    """Softclips or insertions are regarded as residing in the position on the right.
    Therefore, insertion between split_range0.start - 1 and split_range0.start belongs to the split region.
    """
    tuples_before = list()
    for cigartup, target_range0, query_range0 in walk_cigar(cigartuples, reference_start0, with_cigartup=True):
        if len(target_range0) == 0:  # I, S
            if target_range0.start >= split_range0.start:
                break
            else:
                tuples_before.append(cigartup)
        else:  # M, D, N
            if target_range0.start < split_range0.start:
                if target_range0.stop <= split_range0.start:
                    tuples_before.append(cigartup)
                else:
                    truncated_cigarlen = split_range0.start - target_range0.start
                    tuples_before.append((cigartup[0], truncated_cigarlen))

            if target_range0.stop >= split_range0.start:
                break

    return tuples_before


def get_cigars_after_region(cigartuples, reference_start0, reference_end0, split_range0):
    tuples_after = list()
    for cigartup, target_range0, query_range0 in walk_cigar(cigartuples, reference_start0, with_cigartup=True):
        if len(target_range0) == 0:  # I, S
            if target_range0.start < split_range0.stop:
                continue
            else:
                tuples_after.append(cigartup)
        else:  # M, D, N
            if target_range0.stop <= split_range0.stop:
                continue

            if target_range0.start >= split_range0.stop:
                tuples_after.append(cigartup)
            else:
                truncated_cigarlen = target_range0.stop - split_range0.stop
                tuples_after.append((cigartup[0], truncated_cigarlen))

    return tuples_after


def split_read_cigartuples(read, split_range0):
    before_tuples = get_cigars_before_region(read.cigartuples, read.reference_start, read.reference_end, split_range0)
    after_tuples = get_cigars_after_region(read.cigartuples, read.reference_start, read.reference_end, split_range0)

    return before_tuples, after_tuples


def get_cigars_before_region_old(cigartuples, reference_start0, reference_end0, split_range0):
    return _split_cigars_helper(cigartuples, reference_start0, reference_end0, split_range0, include_queryonly_at_border=False)


def get_cigars_after_region_old(cigartuples, reference_start0, reference_end0, split_range0):
    new_cigartuples = cigartuples[::-1]
    new_reference_start0 = 0
    new_reference_end0 = new_reference_start0 + (reference_end0 - reference_start0)
    new_split_range0_start = new_reference_start0 + (reference_end0 - split_range0.stop)
    new_split_range0_end = new_split_range0_start + len(split_range0)
    new_split_range0 = range(new_split_range0_start, new_split_range0_end)

    reversed_result = _split_cigars_helper(new_cigartuples, new_reference_start0, new_reference_end0, new_split_range0, include_queryonly_at_border=True)
    return reversed_result[::-1]


def _split_cigars_helper(cigartuples, reference_start0, reference_end0, split_range0, include_queryonly_at_border):
    """This returns before-region cigartuples

    Insertions or softclips are:
        assigned to the right position in most case
        assigned to the left position when on the left border of the read
    """
    if reference_start0 >= split_range0.start:
        tuples_before = list()
    elif reference_end0 < split_range0.start:
        tuples_before = cigartuples.copy()
    else:
        tuples_before = list()
        current_start0 = reference_start0
        iterator = iter(cigartuples)
        do_further_search = False
        for cigartup in iterator:
            consume_target, consume_query = CIGAR_WALK_DICT[cigartup[0]]
            if consume_target:
                current_start0 += cigartup[1]

            if current_start0 > split_range0.start:
                tuples_before.append(
                    (
                        cigartup[0], 
                        cigartup[1] - (current_start0 - split_range0.start)
                    )
                )
                break
            elif current_start0 == split_range0.start:
                do_further_search = True
                tuples_before.append(cigartup)
                break
            else:
                tuples_before.append(cigartup)

        if do_further_search and include_queryonly_at_border:
            for cigartup in iterator:
                consume_target, consume_query = CIGAR_WALK_DICT[cigartup[0]]
                if (not consume_target) and consume_query:
                    tuples_before.append(cigartup)
                else:
                    break

    return tuples_before


############################################
# Bio.Align.Alignment-related ones #
############################################

def alignment_to_cigartuples(
    alignment, 
    match_as_78=False, del_as_skip=False, 
    left_ins_as_clip=True, right_ins_as_clip=True,
    remove_left_del=True, remove_right_del=True,
):
    """Args:
        alignment: Bio.Align.Alignment object
        assumes global alignment
    """
    # set params
    targetonly_cigarop = 3 if del_as_skip else 2
    match_handler = _match_handler_78 if match_as_78 else _match_handler_0
    # init cigartuples
    cigartuples = list()
    walks = get_walks(alignment, copy=False)
    #list(path_to_walks(alignment.path))
    for target_range0, query_range0 in walks:
        if len(target_range0) == 0:
            cigartuples.append((1, len(query_range0)))
        elif len(query_range0) == 0:
            cigartuples.append((targetonly_cigarop, len(target_range0)))
        else:
            match_handler(alignment, target_range0, query_range0, cigartuples)
    # postprocess - I at borders into clip
    if left_ins_as_clip:
        if cigartuples[0][0] == 1:
            cigartuples[0] = (4, cigartuples[0][1])
    if right_ins_as_clip:
        if cigartuples[-1][0] == 1:
            cigartuples[-1] = (4, cigartuples[-1][1])
    # postprocess - remove flanking targetonly
        # assumes global alignment
    if remove_left_del:
        if cigartuples[0][0] == targetonly_cigarop:
            del cigartuples[0]
    if remove_right_del:
        if cigartuples[-1][0] == targetonly_cigarop:
            del cigartuples[-1]
    # get target offset
    if remove_left_del:
        for target_range0, query_range0 in walks:
            if len(query_range0) > 0:
                target_offset = target_range0.start 
                break
    else:
        target_offset = 0

    return cigartuples, target_offset


def _match_handler_78(alignment, target_range0, query_range0, cigartuples):
    cigarop_list = list()
    for target_idx, query_idx in zip(target_range0, query_range0):
        if alignment.target[target_idx] == alignment.query[query_idx]:
            cigarop_list.append(7)
        else:
            cigarop_list.append(8)

    cigartuples.extend(
        [(x[0], len(tuple(x[1]))) for x in itertools.groupby(cigarop_list)]
    )


def _match_handler_0(alignment, target_range0, query_range0, cigartuples):
    cigartuples.append((0, len(target_range0)))


def walks_to_path(cigarwalks):
    """'path' means 'path' attribute of Bio.Align.PairwiseAlignment object"""
    result = list()
    iterator = iter(cigarwalks)

    target_range0, query_range0 = next(iterator)
    result.append((target_range0.start, query_range0.start))
    result.append((target_range0.stop, query_range0.stop))

    for target_range0, query_range0 in iterator:
        result.append((target_range0.stop, query_range0.stop))

    return tuple(result)


def walks_to_coordinates(walks):
    """Returns what can be used as 'coordinates' attribute of Bio.Align.Alignment object"""
    sublist_target = list()
    sublist_query = list()

    iterator = iter(walks)

    target_range0, query_range0 = next(iterator)
    sublist_target.append(target_range0.start)
    sublist_target.append(target_range0.stop)
    sublist_query.append(query_range0.start)
    sublist_query.append(query_range0.stop)

    for target_range0, query_range0 in iterator:
        sublist_target.append(target_range0.stop)
        sublist_query.append(query_range0.stop)

    return np.array([sublist_target, sublist_query])


def path_to_walks(alignpath):
    """Args:
        alignpath: 'path' attribute of Bio.Align.PairwiseAlignment object
    """
    for x0, x1 in tools.pairwise(alignpath):
        yield range(x0[0], x1[0]), range(x0[1], x1[1])  # target_range, query_range


def coordinates_to_walks(coordinates):
    """Args:
        coordinates: 'coordinates' attribute of Bio.Align.Alignment object
            example: array([[ 0, 59, 72],
                            [ 0, 59, 59]])
    """
    for col_idx in range(coordinates.shape[1] - 1):
        target_range0 = range(coordinates[0, col_idx], coordinates[0, col_idx + 1])
        query_range0 = range(coordinates[1, col_idx], coordinates[1, col_idx + 1])
        yield Walk(target_range0, query_range0)


def set_walks(alignment):
    #alignment.walks = list(path_to_walks(alignment.path))
    alignment.walks = list(coordinates_to_walks(alignment.coordinates))


def get_walks(alignment, copy=False):
    if not hasattr(alignment, 'walks'):
        set_walks(alignment)

    if copy:
        return alignment.walks.copy()
    else:
        return alignment.walks


def amend_outer_insdel_left(alignment):
    walks = get_walks(alignment, copy=False)
    if len(walks) < 3:
        return alignment

    if (
        len(walks[0][0]) == 0 and 
        len(walks[1][1]) == 0 and
        (len(walks[2][0]) > 0 and len(walks[2][1]) > 0)
    ):
        new_walks = walks.copy()
        new_walks[0] = (walks[1][0], range(0, 0))
        new_walks[1] = (range(walks[1][0].stop, walks[1][0].stop), walks[0][1])
        return Bio.Align.Alignment(
            sequences=[alignment.target, alignment.query],
            coordinates=walks_to_coordinates(new_walks),
        )
#        return Bio.Align.PairwiseAlignment(
#            alignment.target, 
#            alignment.query, 
#            walks_to_path(new_walks),
#            0,
#        )
    else:
        return alignment


def amend_outer_insdel_right(alignment):
    walks = get_walks(alignment, copy=False)
    if len(walks) < 3:
        return alignment

    if (
        len(walks[-1][0]) == 0 and 
        len(walks[-2][1]) == 0 and
        (len(walks[-3][0]) > 0 and len(walks[-3][1]) > 0)
    ):
        new_walks = walks.copy()
        new_walks[-1] = (walks[-2][0], range(0, 0))
        new_walks[-2] = (range(walks[-2][0].start, walks[-2][0].start), walks[-1][1])
        return Bio.Align.Alignment(
            sequences=[alignment.target, alignment.query],
            coordinates=walks_to_coordinates(new_walks),
        )
        #return Bio.Align.PairwiseAlignment(
        #    alignment.target, 
        #    alignment.query, 
        #    walks_to_path(new_walks),
        #    0,
        #)
    else:
        return alignment

def amend_outer_insdel_both(alignment):
    return amend_outer_insdel_right(
        amend_outer_insdel_left(alignment)
    )


#def amend_outer_insdel_right(alignment):
#    rev_alignment = reverse_alignment(alignment)
#    rev_alignment_amended = amend_outer_insdel_left(rev_alignment)
#    return reverse_alignment(rev_alignment_amended)


def show_alignment_with_numbers(alignment):
    aln_length = sum(max(len(x), len(y)) for (x, y) in get_walks(alignment))
    string_tens = ''.join(
        str(x) + ' ' * (9 if x == 0 else 9 - np.floor(np.log10(x)).astype('int'))
        for x in range(aln_length // 10 + 1)
    )
    string_ones = ''.join(str(x)[-1] for x in range(aln_length))
    print(string_tens)
    print(string_ones)
    print(alignment, end='')
    print(string_ones)
    print(string_tens)


def reverse_alignment(alignment):
    """Args:
        alignment: Bio.Align.Alignment object
    """
    return Bio.Align.Alignment(
        sequences=[alignment.target[::-1], alignment.query[::-1]],
        coordinates=np.array(
            [
                len(alignment.target) - alignment.coordinates[0][::-1],
                len(alignment.query) - alignment.coordinates[1][::-1],
            ],
        ),
    )
    #return Bio.Align.PairwiseAlignment(
    #    alignment.target[::-1], 
    #    alignment.query[::-1], 
    #    [
    #        (len(alignment.target) - x[0], len(alignment.query) - x[1]) 
    #        for x in alignment.path[::-1]
    #    ], 
    #    0,
    #)


#def remove_flanking_query_gaps(alignment):
#    offset = 0
#    # strip leading query gap
#    walks = list(path_to_walks(alignment.path))
#    for idx, (target_walk, query_walk) in enumerate(walks):
#        if len(query_walk) != 0:
#            break
#        else:
#            offset += len(target_walk)
#    walks = walks[idx:]
#    # correct coordinates of target_walk's
#    for idx in range(len(walks)):
#        old_target_walk, old_query_walk = walks[idx]
#        new_target_walk = range(old_target_walk.start - offset, old_target_walk.stop - offset)
#        walks[idx] = (new_target_walk, old_query_walk)
#    # strip trailing query gap
#    for idx, (target_walk, query_walk) in enumerate(walks[::-1]):
#        if len(query_walk) != 0:
#            break
#    if idx > 0:
#        walks = walks[:-idx]
#    # make new alignment
#    new_aln = Bio.Align.PairwiseAlignment(
#        alignment.target[offset:], 
#        alignment.query,
#        walks_to_path(walks),
#        0,
#    )
#        
#    return new_aln, offset


def remove_leading_query_gaps_component_input(target, query, walks, as_alignment=True):
    offset = 0
    for idx, (target_walk, query_walk) in enumerate(walks):
        if len(query_walk) == 0:
            offset += len(target_walk)
        else:
            break
    new_walks = walks[idx:]
    new_target = target[offset:]
    # correct coordinates of target_walk's
    for idx in range(len(new_walks)):
        old_target_walk, old_query_walk = new_walks[idx]
        new_target_walk = range(old_target_walk.start - offset, old_target_walk.stop - offset)
        new_walks[idx] = (new_target_walk, old_query_walk)
    # return
    if as_alignment:
        #new_aln = Bio.Align.PairwiseAlignment(new_target, query, walks_to_path(new_walks), 0)
        new_aln = Bio.Align.Alignment(
            sequences=[new_target, query],
            coordinates=walks_to_coordinates(new_walks),
        )
        return new_aln, offset
    else:
        return new_target, new_walks, offset


def remove_leading_query_gaps(alignment, as_alignment=True):
    return remove_leading_query_gaps_component_input(alignment.target, alignment.query, get_walks(alignment, copy=False), as_alignment=as_alignment)


def remove_trailing_query_gaps_component_input(target, query, walks, as_alignment=True):
    offset = 0
    for idx, (target_walk, query_walk) in enumerate(walks[::-1]):
        if len(query_walk) == 0:
            offset += len(target_walk)
        else:
            break

    if idx > 0:
        new_walks = walks[:-idx]
    else:
        new_walks = walks.copy()

    if offset == 0:
        new_target = target
    else:
        new_target = target[:-offset]

    # return
    if as_alignment:
        #new_aln = Bio.Align.PairwiseAlignment(new_target, query, walks_to_path(new_walks), 0)
        new_aln = Bio.Align.Alignment(
            sequences=[new_target, query],
            coordinates=walks_to_coordinates(new_walks),
        )
        return new_aln, offset
    else:
        return new_target, new_walks, offset


def remove_trailing_query_gaps(alignment, as_alignment=True):
    return remove_trailing_query_gaps_component_input(alignment.target, alignment.query, get_walks(alignment, copy=False), as_alignment=as_alignment)


def remove_flanking_query_gaps(alignment):
    target_lstrip, walks_lstrip, leading_offset = remove_leading_query_gaps(alignment, as_alignment=False)
    target_lrstrip, walks_lrstrip, trailing_offset = remove_trailing_query_gaps_component_input(
        target_lstrip, alignment.query, walks_lstrip, as_alignment=False,
    )
    #new_aln = Bio.Align.PairwiseAlignment(
    #    target_lrstrip, 
    #    alignment.query, 
    #    walks_to_path(walks_lrstrip),
    #    0,
    #)
    new_aln = Bio.Align.Alignment(
        sequences=[target_lrstrip, alignment.query],
        coordinates=walks_to_coordinates(walks_lrstrip),
    )
        
    return new_aln, leading_offset


def alignment_to_vcfspec(
    alignment, target_start0, chrom, fasta, lstrip_query_gaps=True, rstrip_query_gaps=True,
):
    """May return an empty tuple"""

    def groupkey_walks(x):
        if len(x[0]) == 0 or len(x[1]) == 0:
            return 'indel'
        else:
            return 'match'

    def groupkey_matches(x):
        return x[0] != x[1]

    def handle_single_ins(current_pos0, target_idx, query_idx, query_walk, fasta, chrom, new_aln, vcfspec_list, refver):
        pos = current_pos0
        if target_idx == 0:
            ref = fasta.fetch(chrom, current_pos0 - 1, current_pos0)
        else:
            ref = new_aln.target[target_idx - 1]
        inserted_seq = new_aln.query[query_idx:(query_idx + len(query_walk))]
        alt = ref + inserted_seq
        vcfspec_list.append(libvcfspec.Vcfspec(chrom, pos, ref, (alt,), refver=refver, fasta=fasta))

    def handle_single_del(current_pos0, target_idx, target_walk, fasta, chrom, new_aln, vcfspec_list, refver):
        pos = current_pos0
        if target_idx == 0:
            preceding_base = fasta.fetch(chrom, current_pos0 - 1, current_pos0)
        else:
            preceding_base = new_aln.target[target_idx - 1]
        ref = preceding_base + new_aln.target[target_idx:(target_idx + len(target_walk))]
        alt = ref[0]
        vcfspec_list.append(libvcfspec.Vcfspec(chrom, pos, ref, (alt,), refver=refver, fasta=fasta))

    def handle_consecutive_indels(new_aln, walks_subset, target_idx, query_idx, current_pos0, chrom, fasta, vcfspec_list, refver):
        target_walk_length = sum(len(target_walk) for target_walk, query_walk in walks_subset)
        query_walk_length = sum(len(query_walk) for target_walk, query_walk in walks_subset)
        ref = new_aln.target[target_idx:(target_idx + target_walk_length)]
        alt = new_aln.query[query_idx:(query_idx + query_walk_length)]
        pos = current_pos0 + 1
        vcfspec = libvcfspec.Vcfspec(chrom, pos, ref, (alt,), refver=refver, fasta=fasta)
        #vcfspec = vcfspec.parsimonious()
        vcfspec_list.append(vcfspec)

    def handle_match(new_aln, target_idx, query_idx, target_walk, current_pos0, chrom, fasta, vcfspec_list, refver):
        target_seq = new_aln.target[target_idx:(target_idx + len(target_walk))]
        query_seq = new_aln.query[query_idx:(query_idx + len(target_walk))]
        idx_offset = 0
        for is_mismatch, seq_pairs in itertools.groupby(zip(target_seq, query_seq), key=groupkey_matches):
            seq_pairs = tuple(seq_pairs)
            if is_mismatch:
                pos = current_pos0 + idx_offset + 1
                ref = ''.join(x[0] for x in seq_pairs)
                alt = ''.join(x[1] for x in seq_pairs)
                vcfspec_list.append(libvcfspec.Vcfspec(chrom, pos, ref, (alt,), refver=refver, fasta=fasta))
            idx_offset += len(seq_pairs)

    # set paramters
    refver = refgenome.infer_refver_fasta(fasta)

    if lstrip_query_gaps:
        if rstrip_query_gaps:
            new_aln, leading_offset = remove_flanking_query_gaps(alignment)
        else:
            new_aln, leading_offset = remove_leading_query_gaps(alignment, as_alignment=True)
    else:
        if rstrip_query_gaps:
            new_aln, trailing_offset = remove_trailing_query_gaps(alignment, as_alignment=True)
            leading_offset = 0
        else:
            new_aln = alignment
            leading_offset = 0

    target_start0 += leading_offset
    current_pos0 = target_start0
    target_idx = 0
    query_idx = 0
    vcfspec_list = list()
    
    # main
    for walktype, walks_subset in itertools.groupby(get_walks(new_aln, copy=False), key=groupkey_walks):
        walks_subset = tuple(walks_subset)

        if walktype == 'match':
            if len(walks_subset) != 1:
                raise Exception(
                    f'Longer than one consecutive runs of match walks.\n'
                    f'Input alignment:\n'
                    f'{alignment}'
                )
            target_walk, query_walk = walks_subset[0]
            handle_match(new_aln, target_idx, query_idx, target_walk, current_pos0, chrom, fasta, vcfspec_list, refver)
        elif walktype == 'indel':
            if len(walks_subset) == 1:
                target_walk, query_walk = walks_subset[0]
                if len(target_walk) == 0:  # ins
                    handle_single_ins(current_pos0, target_idx, query_idx, query_walk, fasta, chrom, new_aln, vcfspec_list, refver)
                elif len(query_walk) == 0:  # del
                    handle_single_del(current_pos0, target_idx, target_walk, fasta, chrom, new_aln, vcfspec_list, refver)
            else:
                if not (
                    any(len(target_walk) > 0 for target_walk, query_walk in walks_subset) and 
                    any(len(query_walk) > 0 for target_walk, query_walk in walks_subset)
                ):
                    raise Exception(
                        f'Consecutive run of indels are composed only of insertion or deletion.\n'
                        f'Input alignment:\n'
                        f'{alignment}'
                    )
                handle_consecutive_indels(new_aln, walks_subset, target_idx, query_idx, current_pos0, chrom, fasta, vcfspec_list, refver)

        for target_walk, query_walk in walks_subset:
            current_pos0 += len(target_walk)
            target_idx += len(target_walk)
            query_idx += len(query_walk)
        
    return tuple(vcfspec_list)


def alignment_hasher(aln):
    return (
        tuple(aln.sequences), 
        tuple(map(tuple, aln.coordinates)),
    )


def remove_identical_alignments(alignments):
    hash_store = set()
    for x in alignments:
        hash_val = alignment_hasher(x)
        if hash_val in hash_store:
            continue
        else:
            hash_store.add(hash_val)
            yield x


############################
# alignment post-selection #
############################

class AlignmentTieError(Exception):
    pass


def tiebreaker_scorer_leftmost_gaps(alignment):
    """Those with least "sum of gap start coordinates" are favored"""
    score = 0
    for target_range0, query_range0 in get_walks(alignment, copy=False):
        if len(target_range0) == 0 or len(query_range0) == 0:
            score -= target_range0.start
    return score


def tiebreaker_scorer_gap_length_sum(alignment):
    """Those with shorter gap lengths sum are favored"""
    score = 0
    walks = get_walks(alignment, copy=False)
    for idx, (target_range0, query_range0) in enumerate(walks):
        if len(target_range0) == 0:
            score -= len(query_range0)
        elif len(query_range0) == 0:
            if idx != 0 and idx != (len(walks) - 1):
                score -= len(target_range0)
    return score


def tiebreaker_scorer_gap_ordering(alignment):
    """Favors longer gaps coming earlier than shorter gaps"""
    score = 0
    walks = get_walks(alignment, copy=False)
    for idx, (target_range0, query_range0) in enumerate(walks):
        if len(target_range0) == 0:
            score -= len(query_range0) * target_range0.start
        elif len(query_range0) == 0:
            if idx != 0 and idx != (len(walks) - 1):
                score -= len(target_range0) * query_range0.start
    return score


def tiebreakers_merged_main(alignments):
    # alignment with the highest score is selected

    # step 1
    selected_alns = tools.multi_max(
        alignments, 
        key=tiebreaker_scorer_gap_length_sum, 
        with_target_val=False,
    )
    if len(selected_alns) != 1:
        # step 2
        selected_alns = tools.multi_max(
            selected_alns, 
            key=tiebreaker_scorer_leftmost_gaps, 
            with_target_val=False,
        )
        if len(selected_alns) != 1:
            # step 3
            selected_alns = tools.multi_max(
                selected_alns, 
                key=tiebreaker_scorer_gap_ordering, 
                with_target_val=False,
            )

    return selected_alns


def alignment_tiebreaker(alignments, raise_with_failure=True):
    selected_alns = tiebreakers_merged_main(alignments)
    if len(selected_alns) != 1 and raise_with_failure:
        #alignments_string = list()
        #for x in selected_alns:
        #    alignments_string.append(
        #        f'target: {x.target}\nquery: {x.query}\npath: {x.path}\n{str(x)}'
        #    )
        #alignments_string = '\n'.join(alignments_string)
        alignments_string = '\n'.join(
            [
                f'target: {x.target}\nquery: {x.query}\ncoordinates: {x.coordinates}\n{str(x)}'
                for x in selected_alns
            ]
        )
        raise AlignmentTieError(
            f'Failed to break alignment tie. Alignments are:\n{alignments_string}'
        )
        
    return selected_alns[0]





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun May 19 12:21:06 2019

@author: mishugeb
"""

#from SigProfilerExtractor import subroutines as sub

from cmath import cos
import datetime
import platform

# from torch import sign
from SigProfilerAssignment import decompose_subroutines as sub
from SigProfilerAssignment.DecompositionPlots import PlotDecomposition as plot_decomp
import SigProfilerAssignment

import numpy as np
import pandas as pd
import SigProfilerMatrixGenerator
from SigProfilerMatrixGenerator.scripts import SigProfilerMatrixGeneratorFunc as datadump 
from SigProfilerMatrixGenerator.scripts import CNVMatrixGenerator as scna
import sigProfilerPlotting
import os,sys
from PyPDF2 import PdfFileMerger
import fitz
import time

def convert_PDF_to_PNG(input_file_name, output_directory, page_names):
    pdf_doc = fitz.open(input_file_name)
    zoom = 3
    magnify = fitz.Matrix(zoom, zoom)
    
    if pdf_doc.page_count != len(page_names):
        raise ValueError("Error: The number of samples and number of plots do not match.")
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)
    
    for sample_name, page in zip(page_names, pdf_doc):
        pix = page.get_pixmap(matrix=magnify)
        out_file_name = os.path.join(output_directory, sample_name+  ".png")
        pix.save(out_file_name)

# Create sample reconstruction plots
def generate_sample_reconstruction(cosmic_sigs, samples_input, activities, output_dir, execution_parameters):
    project = "test_run"
    mtype="96"

    final_pdf = PdfFileMerger()
    samples = samples_input.copy(deep=True)
    samples.reset_index(inplace=True)
    for sample_name in samples.columns[1:]:
        # basis names and their corresponding weights
        subset = activities[activities["Samples"].str.contains(sample_name)]
        subset = subset.loc[:, (subset != 0).any(axis=0)]
        basis_names = subset[subset["Samples"].str.contains(sample_name)].columns[1:].tolist()
        recon_tmb = subset.sum(axis=1)
        weights = []
        for i in range(len(basis_names)):
            weights.append(str(float("{0:.6f}".format(100*int(subset[basis_names[i]])/int(recon_tmb))))+"%")

        names_copy = basis_names.copy()
        names_copy.insert(0, cosmic_sigs.columns[0])
        result = plot_decomp.run_PlotSSDecomposition(
                    samples[[samples.columns[0],sample_name]],
                    sample_name,
                    cosmic_sigs[names_copy],
                    basis_names,
                    weights,
                    output_dir,
                    project,
                    mtype,
                    genome_build = execution_parameters["reference_genome"],
                    cosmic_version = str(execution_parameters["cosmic_version"]),
                    exome = execution_parameters["exome"])
        final_pdf.append(result)
    
    pdf_output_path = os.path.join(output_dir,
            "Reconstructed_Sample_Plots_" + str(mtype) + ".pdf")
    web_png_path = os.path.join(output_dir,"WebPNGs")

    final_pdf.write(pdf_output_path)

    convert_PDF_to_PNG(pdf_output_path, web_png_path, samples.columns[1:])
    
    return pdf_output_path

def record_parameters(sysdata,execution_parameters,start_time):
    sysdata.write("\n--------------EXECUTION PARAMETERS--------------\n")
    sysdata.write("INPUT DATA\n")
    sysdata.write("\tinput_type: {}\n".format(execution_parameters["input_type"]))
    sysdata.write("\toutput: {}\n".format(execution_parameters["output"]))
    sysdata.write("\tsamples: {}\n".format(execution_parameters["samples"]))
    sysdata.write("\treference_genome: {}\n".format(execution_parameters["reference_genome"]))
    sysdata.write("\tcontext_types: {}\n".format(execution_parameters["context_type"]))
    sysdata.write("\texome: {}\n".format(execution_parameters["exome"]))

    sysdata.write("COSMIC MATCH\n")
    sysdata.write("\tcosmic_version: {}\n".format(execution_parameters["cosmic_version"]))
    sysdata.write("\tnnls_add_penalty: {}\n".format(execution_parameters["nnls_add_penalty"]))
    sysdata.write("\tnnls_remove_penalty: {}\n".format(execution_parameters["nnls_remove_penalty"]))
    sysdata.write("\tinitial_remove_penalty: {}\n".format(execution_parameters["initial_remove_penalty"]))
    sysdata.write("\tde_novo_fit_penalty: {}\n".format(execution_parameters["de_novo_fit_penalty"]))
    sysdata.write("\texport_probabilities: {}\n".format(execution_parameters["export_probabilities"]))
    sysdata.write("\tcollapse_to_SBS96: {}\n".format(execution_parameters["collapse_to_SBS96"]))
    sysdata.write("\tdenovo_refit_option: {}\n".format(execution_parameters["denovo_refit_option"]))
    sysdata.write("\tdecompose_fit_option: {}\n".format(execution_parameters["decompose_fit_option"]))
    sysdata.write("\tcosmic_fit_option: {}\n".format(execution_parameters["cosmic_fit_option"]))
    sysdata.write("\n-------Analysis Progress------- \n")
    sysdata.write("[{}] Analysis started: \n".format(str(start_time).split(".")[0]))

def spa_analyze(samples, output, input_type='matrix', context_type="96", signatures=None, signature_database=None,decompose_fit_option= True,denovo_refit_option=True,cosmic_fit_option=True, nnls_add_penalty=0.05, 
              nnls_remove_penalty=0.01, initial_remove_penalty=0.05, de_novo_fit_penalty=0.02, 
              genome_build="GRCh37", cosmic_version=3.3, make_plots=True, collapse_to_SBS96=True,connected_sigs=True, verbose=False,devopts=None,new_signature_thresh_hold=0.8,
              exclude_signature_subgroups=None, exome=False, export_probabilities=True, export_probabilities_per_mutation=False, sample_reconstruction_plots=False):

    
    """
    Decomposes the De Novo Signatures into COSMIC Signatures and assigns COSMIC signatures into samples.
    
    Parameters: 
        
        signatures: A string. Path to a  tab delimited file that contains the signaure table where the rows are mutation types and colunms are signature IDs. 
        activities: A string. Path to a tab delimilted file that contains the activity table where the rows are sample IDs and colunms are signature IDs.
        samples: A string. Path to a tab delimilted file that contains the activity table where the rows are mutation types and colunms are sample IDs.
        output: A string. Path to the output folder.
        genome_build = A string. The reference genome build. List of supported genomes: "GRCh37", "GRCh38", "mm9", "mm10" and "rn6". The default value is "GRCh37". If the selected genome is not in the supported list, the default genome will be used.
        verbose = Boolean. Prints statements. Default value is False. 
        exome = Boolean. Defines if the exome renormalized signatures will be used. The default value is False.
        
    Values:
        The files below will be generated in the output folder. 
        
        Cluster_of_Samples.txt
        comparison_with_global_ID_signatures.csv
        Decomposed_Solution_Activities.txt
        Decomposed_Solution_Samples_stats.txt
        Decomposed_Solution_Signatures.txt
        decomposition_logfile.txt
        dendogram.pdf
        Mutation_Probabilities.txt
        Signature_assaignment_logfile.txt
        Signature_plot[MutatutionContext]_plots_Decomposed_Solution.pdf
        
    Example:
        >>>from SigProfilerExtractor import decomposition as decomp
        >>>signatures = "path/to/dDe_Novo_Solution_Signatures.txt"
        >>>activities="path/to/De_Novo_Solution_Activities.txt"
        >>>samples="path/to/Samples.txt"
        >>>output="name or path/to/output.txt"
        decomp.decompose(signatures, activities, samples, output, genome_build="GRCh37", verbose=False)

    """
    if devopts == None:
        layer_directory1 = output+"/De_Novo_Solution"
        layer_directory2 = output+"/Decompose_Solution"
        layer_directory3 = output+"/Assignment_Solution"
    else:
        layer_directory1 = devopts['denovo_outpath']
        layer_directory2 = devopts['decompose_outpath']
        layer_directory3 = devopts['Assignment_outpath']


    if (denovo_refit_option == True or decompose_fit_option ==True) and signatures is None:
         raise Exception("If denovo_refit or decompose_fit is True, signatures cannot be empty")
    
    if input_type=="vcf":
        project_name = 'Input_vcffiles'
        vcf_context = context_type
        data = datadump.SigProfilerMatrixGeneratorFunc(project_name, genome_build, samples, exome=exome, bed_file=None, chrom_based=False, plot=False, gs=False)
        genomes = data[vcf_context]

    elif input_type.split(":")[0].lower()=="seg":
        cnv_file_type = input_type.split(":")[1].upper()
        context_type = "CNV48"
        genomes = scna.generateCNVMatrix(cnv_file_type, samples, cnv_file_type, os.path.dirname(samples) + '/')
        genomes = genomes.set_index('MutationType')

    elif input_type=="matrix":
        try:
            genomes = pd.read_csv(samples, sep = "\t", index_col = 0)
        except:
            genomes = samples
            genomes = pd.DataFrame(genomes)
    else:
        sys.exit("Invalid input_type specified")

    mutation_type = str(genomes.shape[0])
    m=mutation_type

    m_for_subgroups = ''
    if m=='96' or m=='288' or m=='1536':
        m_for_subgroups = 'SBS'
    if m=='78':
        m_for_subgroups = 'DBS'
    if m=='83':
        m_for_subgroups = 'ID'


    default_subgroups_dict= {'remove_MMR_deficiency_signatures' :False,
                      'remove_POL_deficiency_signatures' :False,
                      'remove_HR_deficiency_signatures' :False,
                      'remove_BER_deficiency_signatures' :False,
                      'remove_Chemotherapy_signatures' :False,
                      'remove_Immunosuppressants_signatures' :False,
                      'remove_Treatment_signatures' :False,
                      'remove_APOBEC_signatures' :False,
                      'remove_Tobacco_signatures' :False,
                      'remove_UV_signatures' :False,
                      'remove_AA_signatures' :False,
                      'remove_Colibactin_signatures' :False,
                      'remove_Artifact_signatures' :False,
                      'remove_Lymphoid_signatures' :False}
                      
    default_subgroups_siglists= {'remove_MMR_deficiency_signatures' :{'SBS':['6', '14', '15', '20', '21', '26', '44'], 'DBS':['7', '10'], 'ID':['7']},
                      'remove_POL_deficiency_signatures' :{'SBS':['10a', '10b', '10c', '10d', '28'], 'DBS':['3'], 'ID':[]},
                      'remove_HR_deficiency_signatures' :{'SBS':['3'], 'DBS':[], 'ID':['6']},
                      'remove_BER_deficiency_signatures' :{'SBS':['30','36'], 'DBS':[], 'ID':[]},
                      'remove_Chemotherapy_signatures' :{'SBS':['11','25','31','35','86','87','90'], 'DBS':['5'], 'ID':[]},
                      'remove_Immunosuppressants_signatures' :{'SBS':['32'], 'DBS':[], 'ID':[]},
                      'remove_Treatment_signatures' :{'SBS':['11','25','31','32','35','86','87','90'], 'DBS':['5'], 'ID':[]},
                      'remove_APOBEC_signatures' :{'SBS':['2','13'], 'DBS':[], 'ID':[]},
                      'remove_Tobacco_signatures' :{'SBS':['4','29','92'], 'DBS':['2'], 'ID':['3']},
                      'remove_UV_signatures' :{'SBS':['7a','7b','7c','7d','38'], 'DBS':['1'], 'ID':['13']},
                      'remove_AA_signatures' :{'SBS':['22'], 'DBS':[], 'ID':[]},
                      'remove_Colibactin_signatures' :{'SBS':['88'], 'DBS':[], 'ID':['18']},
                      'remove_Artifact_signatures' :{'SBS':['27','43','45','46','47','48','49','50','51','52','53','54','55','56','57','58','59','60','95'], 'DBS':[], 'ID':[]},
                      'remove_Lymphoid_signatures' :{'SBS':['9','84','85'], 'DBS':[], 'ID':[]}}
    
    
    signature_subgroups_dict = default_subgroups_dict.copy()
    if exclude_signature_subgroups == None:
        pass
    else:
        if type(exclude_signature_subgroups) is not list:
            sys.exit("exclude_signature_subgroups input should be a list of appropriate flags, please refer to documentation.")
        else:

            for key in default_subgroups_dict:
                if key in exclude_signature_subgroups:
                    signature_subgroups_dict[key]=True

    sig_exclusion_list=[]
    if exclude_signature_subgroups == None:
        sig_exclusion_list=[]
    else:
        for key in signature_subgroups_dict:
            if signature_subgroups_dict[key]:
                sig_exclusion_list.append(default_subgroups_siglists[key][m_for_subgroups])

    sig_exclusion_list = [item for sublist in sig_exclusion_list for item in sublist]

    try:
        if not os.path.exists(output):
            os.makedirs(output)
    except:
        print ("The {} folder could not be created".format("output"))

    start_time = datetime.datetime.now() # start time

    # create JOB_METADATA
    sysdata = open(output+"/JOB_METADATA.txt", "w")
    sysdata.write("THIS FILE CONTAINS THE METADATA ABOUT SYSTEM AND RUNTIME\n\n\n")
    sysdata.write("-------System Info-------\n")
    sysdata.write("Operating System Name: "+ platform.uname()[0]+"\n"+"Nodename: "+platform.uname()[1]+"\n"+"Release: "+platform.uname()[2]+"\n"+"Version: "+platform.uname()[3]+"\n")
    sysdata.write("\n-------Python and Package Versions------- \n")
    sysdata.write("Python Version: "+str(platform.sys.version_info.major)+"."+str(platform.sys.version_info.minor)+"."+str(platform.sys.version_info.micro)+"\n")
    sysdata.write("SigProfilerPlotting Version: "+sigProfilerPlotting.__version__+"\n")
    sysdata.write("SigProfilerMatrixGenerator Version: "+SigProfilerMatrixGenerator.__version__+"\n")
    sysdata.write("SigProfilerAssignment Version: "+SigProfilerAssignment.__version__+"\n")
    sysdata.write("Pandas version: "+pd.__version__+"\n")
    sysdata.write("Numpy version: "+np.__version__+"\n")
    execution_parameters = {"input_type":input_type,
                        "context_type":context_type,
                        "signatures":signatures, 
                        "signature_database":signature_database,
                        "denovo_refit_option":denovo_refit_option,
                        "decompose_fit_option":decompose_fit_option,
                        "cosmic_fit_option":cosmic_fit_option, 
                        "output":output, 
                        "samples":samples, 
                        "reference_genome":genome_build, 
                        "cosmic_version":cosmic_version,
                        "context_type":context_type,
                        "exome":exome, 
                        "nnls_add_penalty":nnls_add_penalty,
                        "nnls_remove_penalty":nnls_remove_penalty,
                        "initial_remove_penalty":initial_remove_penalty,
                        "de_novo_fit_penalty":de_novo_fit_penalty,
                        "collapse_to_SBS96":collapse_to_SBS96,
                        "export_probabilities":export_probabilities,
                        "make_plots":make_plots
                        }
    record_parameters(sysdata,execution_parameters,start_time)
    sysdata.close()
    # Add sequence parameter to control the tmbplot y-axis scale
    if exome == True:
        sequence = 'exome'
    else:
        sequence = 'genome'
    

                                                                #################
                                                                # Denovo refiting #
                                                                #################
    
    if denovo_refit_option == True:
        try:
            processAvg = pd.read_csv(signatures,sep='\t', index_col=0)
        except:
            try:
                processAvg=signatures
            except:
                sys.exit("Error in formatting of input signatures, Pass a text file of signatures in the format of denovo signatures")
          
        if devopts == None:
            listOfSignatures = processAvg.columns
            index = genomes.index
            colnames = genomes.columns
        else:
            listOfSignatures=devopts['listOfSignatures']
            index=devopts['index']
            colnames=devopts['colnames']
            genomes = genomes.set_index(index)
            genomes.columns = colnames
            #genomes = genomes.rename_axis("Mutation Types", axis="columns")
       
        #creating list of mutational type to sync with the vcf type input
        if mutation_type == "78":
            mutation_context = "DBS78"
        elif mutation_type == "83":
            mutation_context = "ID83"
        elif mutation_type=="48":
            mutation_context = "CNV48"
        else:
            mutation_context = "SBS"+mutation_type
        try:
            allsigids = processAvg.columns.to_list()
        except:
            allsigids = list(listOfSignatures)
        processAvg = np.array(processAvg)    
        signature_names = sub.make_letter_ids(idlenth = processAvg.shape[1], mtype = mutation_context)
       
        # create the folder for the final solution/ De Novo Solution
        exposureAvg_dummy = pd.DataFrame(np.random.rand(processAvg.shape[1],genomes.shape[1]),index=listOfSignatures,columns=colnames.to_list()).transpose().rename_axis('Samples')
        exposureAvg = exposureAvg_dummy
        exposureAvg.columns=signature_names   

        

        refit_denovo_signatures= True
        init_rem_denovo = 0.0
        
        # make the texts for signature plotting
        #layer_directory1 = output+"/De_Novo_Solution"
        try:
            if not os.path.exists(layer_directory1):
                os.makedirs(layer_directory1)
        except: 
            print ("The {} folder could not be created".format("De_Novo_Solution"))
    
        listOfSignatures = sub.make_letter_ids(idlenth = processAvg.shape[1], mtype=mutation_context)
        genomes = pd.DataFrame(genomes)
        denovo_exposureAvg = np.array(exposureAvg.T)
        print("\n De Novo Fitting .....")
        current_time_start = datetime.datetime.now()
        with open(output+"JOB_METADATA.txt","a") as sysdata:
            sysdata.write("\n De Novo Fitting .....")


        #exposureAvg = sub.make_final_solution(processAvg, genomes, listOfSignatures, layer_directory1, mutation_type, index,\
                   # colnames,denovo_exposureAvg  = denovo_exposureAvg, add_penalty=nnls_add_penalty, remove_penalty=nnls_remove_penalty, initial_remove_penalty=initial_remove_penalty, de_novo_fit_penalty=de_novo_fit_penalty, connected_sigs=connected_sigs, refit_denovo_signatures=refit_denovo_signatures)    
                   # 
        #######
        attribution={}
        for i in allsigids:
            attribution[i]= [i] 
        #only for SBS96
        if mutation_type == "96" or mutation_type=="288" or mutation_type=="1536":        
            background_sigs = sub.get_indeces(list(allsigids), ['SBS1', 'SBS5'])
            # add connected signatures   
            #different_signatures = ss.add_connected_sigs(different_signatures, list(signames))
        #for other contexts
        else:
            background_sigs = []
        exposureAvg_dummy = pd.DataFrame(np.random.rand(processAvg.shape[1],genomes.shape[1]),index=allsigids,columns=colnames.to_list()).transpose().rename_axis('Samples')

        if devopts == None:
            exposureAvg = sub.make_final_solution(processAvg, genomes, allsigids, layer_directory1, mutation_type, index, colnames, 
                                    cosmic_sigs=True, attribution = attribution, denovo_exposureAvg  = exposureAvg_dummy, sequence=sequence, 
                                    background_sigs=background_sigs, verbose=verbose, genome_build=genome_build, 
                                    add_penalty=nnls_add_penalty, remove_penalty=nnls_remove_penalty,
                                    initial_remove_penalty=init_rem_denovo,connected_sigs=connected_sigs,refit_denovo_signatures=False,
                                    make_plots=make_plots,export_probabilities=export_probabilities, export_probabilities_per_mutation=export_probabilities_per_mutation,
                                    samples=samples, input_type=input_type, denovo_refit_option=denovo_refit_option)

        else:
            signature_stabilities=devopts['signature_stabilities']
            signature_total_mutations=devopts['signature_total_mutations']
            signature_stats = devopts['signature_stats']
            sequence=devopts['sequence']
            processSTE=devopts['processSTE']

            exposureAvg = sub.make_final_solution(processAvg, genomes, allsigids, layer_directory1, mutation_type, index, colnames, 
                                    cosmic_sigs=True, attribution = attribution, denovo_exposureAvg  = exposureAvg_dummy, sequence=sequence, 
                                    background_sigs=background_sigs, verbose=verbose, genome_build=genome_build, signature_total_mutations = signature_total_mutations,
                                    add_penalty=nnls_add_penalty, remove_penalty=nnls_remove_penalty, process_std_error = processSTE, signature_stabilities = signature_stabilities,
                                    initial_remove_penalty=init_rem_denovo,connected_sigs=connected_sigs,refit_denovo_signatures=True,export_probabilities=export_probabilities,
                                    export_probabilities_per_mutation=export_probabilities_per_mutation, samples=samples, input_type=input_type, denovo_refit_option=denovo_refit_option)
        
        with open(output+"JOB_METADATA.txt","a") as sysdata:
            current_time_end = datetime.datetime.now()

            sysdata.write(f"\n Finished Denovo fitting! \nExecution time:{str(current_time_end-current_time_start)}\n")
        

                                                                #################
                                                                # Decomposition       
                                                                #################
    if decompose_fit_option ==True:
        try:
            processAvg = pd.read_csv(signatures,sep='\t', index_col=0)
        except:
            try:
                processAvg=signatures
            except:
                sys.exit("Error in formatting of input signatures, Pass a text file of signatures in the format of denovo signatures")

        if devopts == None:
            listOfSignatures = processAvg.columns
            index = genomes.index
            colnames = genomes.columns
            make_decomposition_plots=make_plots
        else:
            listOfSignatures=devopts['listOfSignatures']
            index=devopts['index']
            colnames=devopts['colnames']
            genomes = genomes.set_index(index)
            genomes.columns = colnames
            make_decomposition_plots = devopts['make_decomposition_plots']
            #genomes = genomes.rename_axis("Mutation Types", axis="columns")

        #creating list of mutational type to sync with the vcf type input
        if mutation_type == "78":
            mutation_context = "DBS78"
        elif mutation_type == "83":
            mutation_context = "ID83"
        elif mutation_type=="48":
            mutation_context = "CNV48"
        else:
            mutation_context = "SBS"+mutation_type
        try:
            allsigids = processAvg.columns.to_list()
        except:
            allsigids = list(listOfSignatures)
        processAvg = np.array(processAvg)    
        signature_names = sub.make_letter_ids(idlenth = processAvg.shape[1], mtype = mutation_context)

        exposureAvg_dummy = pd.DataFrame(np.random.rand(processAvg.shape[1],genomes.shape[1]),index=listOfSignatures,columns=colnames.to_list()).transpose().rename_axis('Samples')
        exposureAvg = exposureAvg_dummy
        exposureAvg.columns=signature_names   


        #############################
        #layer_directory2 = output+"/Decompose_Solution"
        if isinstance(processAvg, pd.DataFrame):
            pass
        else:
            originalProcessAvg=pd.DataFrame(processAvg,index=index,columns=listOfSignatures)
        try:
            if not os.path.exists(layer_directory2):
                os.makedirs(layer_directory2)
        except: 
            print ("The {} folder could not be created".format("Decomposed_Solution"))

        if processAvg.shape[0]==1536 and collapse_to_SBS96==True: #collapse the 1596 context into 96 only for the deocmposition 
            processAvg = pd.DataFrame(processAvg, index=index)
            processAvg = processAvg.groupby(processAvg.index.str[1:8]).sum()
            genomes = genomes.groupby(genomes.index.str[1:8]).sum()
            index = genomes.index
            processAvg = np.array(processAvg)
        
        if processAvg.shape[0]==288 and collapse_to_SBS96==True: #collapse the 288 context into 96 only for the deocmposition 
            processAvg = pd.DataFrame(processAvg, index=index)
            processAvg = processAvg.groupby(processAvg.index.str[2:9]).sum()
            genomes = pd.DataFrame(genomes, index=index)
            genomes = genomes.groupby(genomes.index.str[2:9]).sum()
            index = genomes.index
            processAvg = np.array(processAvg)
                
        print("\n Decomposing De Novo Signatures  .....")
        current_time_start = datetime.datetime.now()
        with open(output+"JOB_METADATA.txt","a") as sysdata:
            sysdata.write("\n Decomposing De Novo Signatures  .....")
        final_signatures = sub.signature_decomposition(processAvg, mutation_type, layer_directory2, genome_build=genome_build,cosmic_version=cosmic_version,signature_database=signature_database, mutation_context=mutation_context, add_penalty=0.05, connected_sigs=connected_sigs,remove_penalty=0.01, make_decomposition_plots=make_decomposition_plots, originalProcessAvg=originalProcessAvg,new_signature_thresh_hold=new_signature_thresh_hold,sig_exclusion_list=sig_exclusion_list,exome=exome, m_for_subgroups=m_for_subgroups)    
        #final_signatures = sub.signature_decomposition(processAvg, m, layer_directory2, genome_build=genome_build)
        # extract the global signatures and new signatures from the final_signatures dictionary
        globalsigs = final_signatures["globalsigs"]
        globalsigs = np.array(globalsigs)
        newsigs = final_signatures["newsigs"]
        processAvg = np.hstack([globalsigs, newsigs])  
        allsigids = final_signatures["globalsigids"]+final_signatures["newsigids"]
        attribution = final_signatures["dictionary"]
        background_sigs= final_signatures["background_sigs"]
        index = genomes.index
        colnames = genomes.columns

        print("\n Assigning decomposed signature")
        with open(output+"JOB_METADATA.txt","a") as sysdata:
            sysdata.write("\n [{}] Assigning decomposed signature \n".format(str(start_time).split(".")[0]))
            
        result = sub.make_final_solution(processAvg, genomes, allsigids, layer_directory2, mutation_type, index, colnames, 
                                cosmic_sigs=True, attribution = attribution, denovo_exposureAvg  = exposureAvg, sequence=sequence,  
                                background_sigs=background_sigs, verbose=verbose, genome_build=genome_build, 
                                add_penalty=nnls_add_penalty, remove_penalty=nnls_remove_penalty, 
                                initial_remove_penalty=initial_remove_penalty,connected_sigs=connected_sigs,
                                collapse_to_SBS96=collapse_to_SBS96,
                                refit_denovo_signatures=False,
                                make_plots=make_plots,
                                export_probabilities=export_probabilities,
                                export_probabilities_per_mutation=export_probabilities_per_mutation,
                                samples=samples, input_type=input_type, denovo_refit_option=denovo_refit_option)

        with open(output+"JOB_METADATA.txt","a") as sysdata:
            current_time_end = datetime.datetime.now()

            sysdata.write(f"\n Finished Decompose fitting! \nExecution time:{str(current_time_end-current_time_start)}\n")
                                                                #################
                                                                # Cosmic Fitting       
                                                                #################
        
    if cosmic_fit_option ==True:
        try:
            if not os.path.exists(layer_directory3):
                os.makedirs(layer_directory3)
        except: 
            print ("The {} folder could not be created".format("Assignment_Solution"))
        # if signatures == None:
        #     processAvg = sub.getProcessAvg(genomes, genome_build, "3.2")
        #     processAvg = processAvg.set_index('Type').rename_axis('MutationType')
        # else:
        #     try:
        #         processAvg = pd.read_csv(signatures,sep='\t', index_col=0)
        #     except:
        #         sys.exit("Something is wrong with the format of input signatures, Pass a text file of signatures in the format of COSMIC sig database")
        index = genomes.index
        colnames = genomes.columns
        if (genomes.sum()==0).sum() >0:
            print("Removing samples with zero TMB ...... ")
            genomes=genomes.loc[:, (genomes != 0).any(axis=0)]
            colnames = genomes.columns
        
        if genomes.shape[0]==1536 and collapse_to_SBS96==True: #collapse the 1596 context into 96 only for the deocmposition 
            # processAvg = pd.DataFrame(processAvg, index=index)
            # processAvg = processAvg.groupby(processAvg.index.str[1:8]).sum()
            genomes = pd.DataFrame(genomes, index=index)
            genomes = genomes.groupby(genomes.index.str[1:8]).sum()
            index = genomes.index
            #processAvg = np.array(processAvg)
        
        if genomes.shape[0]==288 and collapse_to_SBS96==True: #collapse the 288 context into 96 only for the deocmposition 
            # processAvg = pd.DataFrame(processAvg, index=index)
            # processAvg = processAvg.groupby(processAvg.index.str[2:9]).sum()
            genomes = pd.DataFrame(genomes, index=index)
            genomes = genomes.groupby(genomes.index.str[2:9]).sum()
            index = genomes.index


        if signature_database==None:
            processAvg = sub.getProcessAvg(genomes, genome_build=genome_build, cosmic_version=cosmic_version, exome=exome)[0]
            # for sample reconstruction plots
            cosmic_sig_ref = processAvg.copy(deep=True)
            cosmic_sig_ref.reset_index(inplace=True)
        else:
            try:
                processAvg = pd.read_csv(signature_database,sep='\t', index_col=0)
            except:
                sys.exit("Something is wrong with the format of signature database, Pass a text file of signatures in the format of COSMIC sig database")
        
        # collapse the 1596 context into 96 only for the decomposition
        if processAvg.shape[0]==1536 and collapse_to_SBS96==True:
            processAvg = pd.DataFrame(processAvg, index=index)
            processAvg = processAvg.groupby(processAvg.index.str[1:8]).sum()
            genomes = pd.DataFrame(genomes, index=index)
            genomes = genomes.groupby(genomes.index.str[1:8]).sum()
            index = genomes.index
        
        if processAvg.shape[0]==288 and collapse_to_SBS96==True: #collapse the 288 context into 96 only for the deocmposition 
            processAvg = pd.DataFrame(processAvg, index=index)
            processAvg = processAvg.groupby(processAvg.index.str[2:9]).sum()
            genomes = pd.DataFrame(genomes, index=index)
            genomes = genomes.groupby(genomes.index.str[2:9]).sum()
            index = genomes.index
   

        #processAvg is sigdatabase: remove sigs corresponding to exclusion rules.
        sig_exclusion_list= [m_for_subgroups+items for items in sig_exclusion_list]
        if sig_exclusion_list:
            print("The following signatures are excluded: "+" ".join(str(item) for item in sig_exclusion_list))
        # # 
        processAvg.drop(sig_exclusion_list, axis=1, inplace=True,errors='ignore')

        #processAvg= originalProcessAvg
        #index = genomes.index
        #colnames = genomes.columns
        allsigids = processAvg.columns.to_list()
        processAvg = processAvg.values
        attribution={}
        for i in allsigids:
            attribution[i]= [i] 
        #only for SBS96
        if mutation_type == "96" or mutation_type=="288" or mutation_type=="1536":        
            background_sigs = sub.get_indeces(list(allsigids), ['SBS1', 'SBS5'])
            # add connected signatures   
            #different_signatures = ss.add_connected_sigs(different_signatures, list(signames))
        #for other contexts
        else:
            background_sigs = []
        exposureAvg_dummy = pd.DataFrame(np.random.rand(processAvg.shape[1],genomes.shape[1]),index=allsigids,columns=colnames.to_list()).transpose().rename_axis('Samples')
        print("Assigning COSMIC sigs or Signature Database ...... ")
        current_time_start = datetime.datetime.now()
        with open(output+"JOB_METADATA.txt","a") as sysdata:
            sysdata.write("\n Assigning COSMIC sigs or Signature Database ...... ")
        if processAvg.shape[0] != 96:
            if genomes.shape[0] == processAvg.shape[0] and collapse_to_SBS96 ==True:
                sys.exit("Signatures Database and Samples are of same context type and is not equal to 96. please rerun by setting the flag \"collapse_to_SBS96 = False \"")
        


        sub.make_final_solution(processAvg, genomes, allsigids, layer_directory3, mutation_type, index, colnames, 
                            cosmic_sigs=True, attribution = attribution, denovo_exposureAvg  = exposureAvg_dummy, sequence=sequence, 
                            background_sigs=background_sigs, verbose=verbose, genome_build=genome_build, 
                            add_penalty=nnls_add_penalty, remove_penalty=nnls_remove_penalty, 
                            initial_remove_penalty=initial_remove_penalty,connected_sigs=connected_sigs,
                            collapse_to_SBS96=collapse_to_SBS96,refit_denovo_signatures=False,
                            make_plots =make_plots,
                            export_probabilities=export_probabilities,
                            export_probabilities_per_mutation=export_probabilities_per_mutation,
                            samples=samples, input_type=input_type, denovo_refit_option=denovo_refit_option)
        
        with open(output+"JOB_METADATA.txt","a") as sysdata:
            current_time_end = datetime.datetime.now()

            sysdata.write(f"\n Finished Cosmic fitting! \nExecution time:{str(current_time_end-current_time_start)}\n")
    
    # Generate sample reconstruction plots
    if sample_reconstruction_plots and context_type == "96":
        ss_recon_odir = os.path.join(
                    layer_directory3,
                    "Activities",
                    "SampleReconstruction")
        if not os.path.exists(ss_recon_odir):
            os.makedirs(ss_recon_odir)
        acts_path = os.path.join(
                layer_directory3,
                "Activities",
                "Assignment_Solution_Activities.txt")
        cosmic_activities = pd.read_csv(acts_path, sep="\t")
        # Create sample reconstruction plots
        generate_sample_reconstruction(cosmic_sig_ref,
                                        genomes,
                                        cosmic_activities,
                                        ss_recon_odir,
                                        execution_parameters)

    # Complete JOB_METADATA
    sysdata = open(output+"/JOB_METADATA.txt", "a")
    end_time = datetime.datetime.now()
    sysdata.write("\n[{}] Analysis ended: \n".format(str(end_time).split(".")[0]))
    sysdata.write("\n-------Job Status------- \n")
    sysdata.write("Assignment of mutational signatures completed successfully! \nTotal execution time: "+str(end_time-start_time).split(".")[0]+" \nResults can be found in: "+" "+output+ " " +" folder")
    sysdata.close()

    print("\n\n \nYour Job Is Successfully Completed! Thank You For Using SigProfilerAssignment.\n ")
  





import numpy as np
import pyranges as pr

from handygenome.interval import Interval
import handygenome.annotation.ensembl_rest as ensembl_rest
import handygenome.annotation.data as annotdata


def get_gene_coords(gene_names, refver):
    result = dict()
    rest_result = ensembl_rest.lookup_symbol_post(gene_names, refver, expand=False)
    for x in rest_result:
        gene_name = x['display_name']
        chrom = x['seq_region_name']
        start1 = x['start']
        end1 = x['end']
        result[gene_name] = Interval(chrom=chrom, start1=start1, end1=end1)

    return result


def get_gene_coords_gr(gene_names, refver):
    """Genes not searchable from "Ensembl REST lookup symbol" are discarded
    """
    gene_names = list(gene_names)
    data = get_gene_coords(gene_names, refver)
    chroms = list()
    starts = list()
    ends = list()
    valid_genes = list()

    for gene in gene_names:
        if gene in data.keys():
            intv = data[gene]
            chroms.append(intv.chrom)
            starts.append(intv.start0)
            ends.append(intv.end0)
            valid_genes.append(gene)
        else:
            continue

    return pr.from_dict({
        'Chromosome': chroms,
        'Start': starts,
        'End': ends,
        'gene_name': valid_genes,
    })


def get_gene_exon_coords(gene_names, refver):
    assert isinstance(gene_names, (list, tuple))

    result = dict()
    rest_result = ensembl_rest.lookup_symbol_post(gene_names, refver, expand=True)
    for x in rest_result:
            
        #chrom = x['seq_region_name']
        #start0 = x['start'] - 1
        #end0 = x['end']

        gene_name = x['display_name']
        result[gene_name] = dict()
        result[gene_name]['chrom'] = x['seq_region_name']
        result[gene_name]['gene'] = range(x['start'] - 1, x['end'])
        if x['strand'] not in (1, -1):
            raise Exception(f'"strand" value from REST lookup result is neither 1 nor -1.')
        result[gene_name]['is_forward'] = (x['strand'] == 1)

        transcript_info = parse_transcript_info_from_lookup_symbol(x)
        for key, val in transcript_info.items():
            if val['is_canonical']:
                result[gene_name]['exon'] = val['exon'].copy()

    return result


def parse_transcript_info_from_lookup_symbol(lookup_result):
    result = dict()
    for x in lookup_result['Transcript']:
        transcript_id = x['id']
        result[transcript_id] = dict()

        result[transcript_id]['is_canonical'] = (x['is_canonical'] == 1)
        result[transcript_id]['display_name'] = x['display_name']
        result[transcript_id]['exon'] = list()
        for y in x['Exon']:
            start0 = y['start'] - 1
            end0 = y['end']
            result[transcript_id]['exon'].append(range(start0, end0))

    return result


def get_geneset_gr(refver):
    return pr.read_gff3(annotdata.GENESET_PATH[refver])


def get_gene_exon_coords_fromgff(gene_names, geneset_gr, refver=None, rest_result=None):
    """Returns CDS and UTRs as well as exons"""
    if rest_result is None:
        rest_result = ensembl_rest.lookup_symbol_post(gene_names, refver, expand=True)

    result = dict()
    for gene, rest_data in zip(gene_names, rest_result):
        subresult = dict()
        # chrom, gene, is_forward
        subresult = dict()
        subresult['chrom'] = rest_data['seq_region_name']
        subresult['gene'] = range(rest_data['start'] - 1, rest_data['end'])
        if rest_data['strand'] not in (1, -1):
            raise Exception(f'"strand" value from REST lookup result is neither 1 nor -1.')
        subresult['is_forward'] = (rest_data['strand'] == 1)
        subresult['canonical_transcript'] = rest_data['canonical_transcript'].split('.')[0]

        # exon, CDS, UTR
        subgr = geneset_gr[subresult['chrom'], subresult['gene'].start:subresult['gene'].stop]
        subgr = subgr[
            subgr.Parent == f'transcript:{subresult["canonical_transcript"]}'
        ]
        sources = subgr.Source.unique()
        if len(sources) > 1:
            source = max(sources, key=(lambda x: subgr[subgr.Source == x].df.shape[0]))
        else:
            source = sources[0]

        subgr = subgr[subgr.Source == source]
        for key, subdf in subgr.df.groupby('Feature'):
            subresult[key] = list(
                subdf.loc[:, ['Start', 'End']].apply(lambda x: range(*x), axis=1)
            )

        result[gene] = subresult

    return result



        





#!/usr/bin/env python3
 
#Author: Erik Bergstrom

#Contact: ebergstr@eng.ucsd.edu

from __future__ import print_function
import pickle
import os
import argparse
import re

def save_chrom_tsb_separate (genome, ref_dir, custom):
	'''
	Saves the transcriptional strand bias information for a given genome 
	into a BED file.

	Parameters:
		 genome  -> reference genome
		ref_dir  -> directory from which the parent script is being called.

	Returns:
		  None
	
	Outputs:
		BED files that contain the TSB information and their associated ranges.
	'''

	# Instantiates all of the relevant path and reference varibales
	chromosome_path = ref_dir + '/references/chromosomes/tsb/' + genome + '/'
	chromosome_BED_path = ref_dir + '/references/chromosomes/tsb_BED/' + genome + '/'
	chromosomes = ['X', 'Y', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', 
				   '13', '14', '15', '16', '17', '18', '19', '20', '21', '22']
	chromosomes_elegans = ['X','I','II','III','IV','V','MtDNA']
	chromosomes_dog = ['X', 'Y', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', 
				   	   '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', '23','24','25',
				   	   '26','27','28','29','30','31','32','33','34','35','36','37','38','M']
	chromosomes_yeast = ['I','II','III','IV','V','VI','VII','VIII','IX','X','XI','XII','XIII','XIV','XV','XVI']

	tsb_ref = {0:['N','A'], 1:['N','C'], 2:['N','G'], 3:['N','T'],
			   4:['T','A'], 5:['T','C'], 6:['T','G'], 7:['T','T'],
			   8:['U','A'], 9:['U','C'], 10:['U','G'], 11:['U','T'],
			   12:['B','A'], 13:['B','C'], 14:['B','G'], 15:['B','T'],
			   16:['N','N'], 17:['T','N'], 18:['U','N'], 19:['B','N']}
	tsb_ref_original = {'N':0, 'T':1, 'U':2, 'B':3}

	if custom:
		tsb_list = [f for f in os.listdir(chromosome_path) if not f.startswith('.')]
		chromosomes = sorted([tmp_chrom.replace(".txt", "") for tmp_chrom in tsb_list])

	# Truncates the chromosome variable if a mouse genome is provided
	if genome == 'mm10' or genome == 'mm9' or genome == 'mm39':
		chromosomes = chromosomes[:21]

	if genome == 'c_elegans':
		chromosomes = chromosomes_elegans

	if genome == 'dog':
		chromosomes = chromosomes_dog

	if genome == 'yeast':
		chromosomes = chromosomes_yeast

	# Creates the output path if it does not exist
	if not os.path.exists(chromosome_BED_path):
		os.makedirs(chromosome_BED_path)

	# Iterates through each chromosome, saving a new BED file for each
	for chrom in chromosomes:
		try:
			with open (chromosome_path+chrom+".txt", "rb") as f, open(chromosome_BED_path + chrom + "_BED_TSB.txt", 'w') as out:
				print("<CHROM>\t<START>\t<END>\t<TSB>", file=out)
				chrom_tsb = f.read()
				first_tsb = tsb_ref_original[tsb_ref[chrom_tsb[0]][0]]
				current_range = [0]
				for i in range (1, len(chrom_tsb), 1):
					if tsb_ref_original[tsb_ref[chrom_tsb[i]][0]] != first_tsb:
						current_range.append(i-1)
						print(chrom + "\t" + str(current_range[0]) + "\t" + str(current_range[1]) + "\t" + str(tsb_ref_original[tsb_ref[chrom_tsb[i-1]][0]]), file=out)
						first_tsb = tsb_ref_original[tsb_ref[chrom_tsb[i]][0]]
						current_range = [i]
					else:
						continue
			print("chromosome ", chrom, "done")
		except:
			print("chromosome ", chrom, "file not found for this species.")


def main ():
	parser = argparse.ArgumentParser(description="Provide the necessary arguments to install the reference files.")
	parser.add_argument("-g", "--genome", nargs='?', help="Optional parameter instructs script to install the custom genome.")
	parser.add_argument("-c", "--custom", help="custom genome?. (True or False)", default=False)
	args = parser.parse_args()
	genome = args.genome
	custom = bool(args.custom)

	current_dir = os.getcwd()
	ref_dir = re.sub('\/scripts$', '', current_dir)

	save_chrom_tsb_separate(genome, ref_dir, custom)


if __name__ == '__main__':
	main()